@@ -47,8 +47,11 @@ from . import speechserver
 from . import settings
 from . import orca_state
 from . import punctuation_settings
+from . import settings_manager
 from .acss import ACSS
 
+_settingsManager = settings_manager.getManager()
+
 try:
     import speechd
 except:
@@ -62,8 +65,8 @@ else:
     else:
         _speechd_version_ok = True
 
-PUNCTUATION = re.compile('[^\w\s]', re.UNICODE)
-ELLIPSIS = re.compile('(\342\200\246|\.\.\.\s*)')
+PUNCTUATION = re.compile(r'[^\w\s]', re.UNICODE)
+ELLIPSIS = re.compile('(\342\200\246|(?<!\\.)\\.{3,4}(?=(\\s|\\Z)))')
 
 class SpeechServer(speechserver.SpeechServer):
     # See the parent class for documentation.
@@ -73,10 +76,11 @@ class SpeechServer(speechserver.SpeechServer):
     DEFAULT_SERVER_ID = 'default'
     _SERVER_NAMES = {DEFAULT_SERVER_ID: guilabels.DEFAULT_SYNTHESIZER}
 
+    @staticmethod
     def getFactoryName():
         return guilabels.SPEECH_DISPATCHER
-    getFactoryName = staticmethod(getFactoryName)
 
+    @staticmethod
     def getSpeechServers():
         servers = []
         default = SpeechServer._getSpeechServer(SpeechServer.DEFAULT_SERVER_ID)
@@ -85,8 +89,8 @@ class SpeechServer(speechserver.SpeechServer):
             for module in default.list_output_modules():
                 servers.append(SpeechServer._getSpeechServer(module))
         return servers
-    getSpeechServers = staticmethod(getSpeechServers)
 
+    @classmethod
     def _getSpeechServer(cls, serverId):
         """Return an active server for given id.
 
@@ -96,23 +100,20 @@ class SpeechServer(speechserver.SpeechServer):
         """
         if serverId not in cls._active_servers:
             cls(serverId)
-        # Don't return the instance, unless it is succesfully added
+        # Don't return the instance, unless it is successfully added
         # to `_active_Servers'.
         return cls._active_servers.get(serverId)
-    _getSpeechServer = classmethod(_getSpeechServer)
 
+    @staticmethod
     def getSpeechServer(info=None):
-        if info is not None:
-            thisId = info[1]
-        else:
-            thisId = SpeechServer.DEFAULT_SERVER_ID
+        thisId = info[1] if info is not None else SpeechServer.DEFAULT_SERVER_ID
         return SpeechServer._getSpeechServer(thisId)
-    getSpeechServer = staticmethod(getSpeechServer)
 
+    @staticmethod
     def shutdownActiveServers():
-        for server in list(SpeechServer._active_servers.values()):
+        servers = [s for s in SpeechServer._active_servers.values()]
+        for server in servers:
             server.shutdown()
-    shutdownActiveServers = staticmethod(shutdownActiveServers)
 
     # *** Instance methods ***
 
@@ -128,18 +129,22 @@ class SpeechServer(speechserver.SpeechServer):
             (ACSS.FAMILY, self._set_family),
             )
         if not _speechd_available:
-            debug.println(debug.LEVEL_WARNING,
-                          "Speech Dispatcher interface not installed.")
+            msg = 'ERROR: Speech Dispatcher is not available'
+            debug.println(debug.LEVEL_WARNING, msg, True)
             return
         if not _speechd_version_ok:
-            debug.println(debug.LEVEL_WARNING,
-                        "Speech Dispatcher version 0.6.2 or later is required.")
+            msg = 'ERROR: Speech Dispatcher version 0.6.2 or later is required.'
+            debug.println(debug.LEVEL_WARNING, msg, True)
             return
         # The following constants must be initialized in runtime since they
         # depend on the speechd module being available.
+        try:
+            most = speechd.PunctuationMode.MOST
+        except:
+            most = speechd.PunctuationMode.SOME
         self._PUNCTUATION_MODE_MAP = {
             settings.PUNCTUATION_STYLE_ALL:  speechd.PunctuationMode.ALL,
-            settings.PUNCTUATION_STYLE_MOST: speechd.PunctuationMode.SOME,
+            settings.PUNCTUATION_STYLE_MOST: most,
             settings.PUNCTUATION_STYLE_SOME: speechd.PunctuationMode.SOME,
             settings.PUNCTUATION_STYLE_NONE: speechd.PunctuationMode.NONE,
             }
@@ -147,7 +152,7 @@ class SpeechServer(speechserver.SpeechServer):
             speechd.CallbackType.BEGIN: speechserver.SayAllContext.PROGRESS,
             speechd.CallbackType.CANCEL: speechserver.SayAllContext.INTERRUPTED,
             speechd.CallbackType.END: speechserver.SayAllContext.COMPLETED,
-           #speechd.CallbackType.INDEX_MARK:speechserver.SayAllContext.PROGRESS,
+            speechd.CallbackType.INDEX_MARK:speechserver.SayAllContext.PROGRESS,
             }
 
         self._default_voice_name = guilabels.SPEECH_DEFAULT_VOICE % serverId
@@ -155,9 +160,9 @@ class SpeechServer(speechserver.SpeechServer):
         try:
             self._init()
         except:
-            debug.println(debug.LEVEL_WARNING,
-                          "Speech Dispatcher service failed to connect:")
             debug.printException(debug.LEVEL_WARNING)
+            msg = 'ERROR: Speech Dispatcher service failed to connect'
+            debug.println(debug.LEVEL_WARNING, msg, True)
         else:
             SpeechServer._active_servers[serverId] = self
 
@@ -171,10 +176,27 @@ class SpeechServer(speechserver.SpeechServer):
         self._current_voice_properties = {}
         mode = self._PUNCTUATION_MODE_MAP[settings.verbalizePunctuationStyle]
         client.set_punctuation(mode)
+        client.set_data_mode(speechd.DataMode.SSML)
 
     def updateCapitalizationStyle(self):
         """Updates the capitalization style used by the speech server."""
-        self._client.set_cap_let_recogn(settings.capitalizationStyle)
+
+        if settings.capitalizationStyle == settings.CAPITALIZATION_STYLE_ICON:
+            style = 'icon'
+        elif settings.capitalizationStyle == settings.CAPITALIZATION_STYLE_SPELL:
+            style = 'spell'
+        else:
+            style = 'none'
+
+        try:
+            self._client.set_cap_let_recogn(style)
+        except speechd.SSIPCommunicationError:
+            msg = "SPEECH DISPATCHER: Connection lost. Trying to reconnect."
+            debug.println(debug.LEVEL_INFO, msg, True)
+            self.reset()
+            self._client.set_cap_let_recogn(style)
+        except:
+            pass
 
     def updatePunctuationLevel(self):
         """ Punctuation level changed, inform this speechServer. """
@@ -182,20 +204,15 @@ class SpeechServer(speechserver.SpeechServer):
         self._client.set_punctuation(mode)
 
     def _send_command(self, command, *args, **kwargs):
-        if hasattr(speechd, 'SSIPCommunicationError'):
-            try:
-                return command(*args, **kwargs)
-            except speechd.SSIPCommunicationError:
-                debug.println(debug.LEVEL_CONFIGURATION,
-                              "Speech Dispatcher connection lost. "
-                              "Trying to reconnect.")
-                self.reset()
-                return command(*args, **kwargs)
-            except:
-                pass
-        else:
-            # It is not possible tho catch the error with older SD versions. 
+        try:
             return command(*args, **kwargs)
+        except speechd.SSIPCommunicationError:
+            msg = "SPEECH DISPATCHER: Connection lost. Trying to reconnect."
+            debug.println(debug.LEVEL_INFO, msg, True)
+            self.reset()
+            return command(*args, **kwargs)
+        except:
+            pass
 
     def _set_rate(self, acss_rate):
         rate = int(2 * max(0, min(99, acss_rate)) - 98)
@@ -209,15 +226,33 @@ class SpeechServer(speechserver.SpeechServer):
         volume = int(15 * max(0, min(9, acss_volume)) - 35)
         self._send_command(self._client.set_volume, volume)
 
-    def _set_family(self, acss_family):
-        familyLocale = acss_family.get(speechserver.VoiceFamily.LOCALE)
-        if not familyLocale:
+    def _get_language_and_dialect(self, acss_family):
+        if acss_family is None:
+            acss_family = {}
+
+        language = acss_family.get(speechserver.VoiceFamily.LANG)
+        dialect = acss_family.get(speechserver.VoiceFamily.DIALECT)
+
+        if not language:
             import locale
             familyLocale, encoding = locale.getdefaultlocale()
-        if familyLocale:
-            lang = familyLocale.split('_')[0]
-            if lang and len(lang) == 2:
-                self._send_command(self._client.set_language, str(lang))
+
+            language, dialect = '', ''
+            if familyLocale:
+                localeValues = familyLocale.split('_')
+                language = localeValues[0]
+                if len(localeValues) == 2:
+                    dialect = localeValues[1]
+
+        return language, dialect
+
+    def _set_family(self, acss_family):
+        lang, dialect = self._get_language_and_dialect(acss_family)
+        self._send_command(self._client.set_language, lang)
+        if dialect:
+            # Try to set precise dialect
+            self._send_command(self._client.set_language, lang + '-' + dialect)
+
         try:
             # This command is not available with older SD versions.
             set_synthesis_voice = self._client.set_synthesis_voice
@@ -225,9 +260,43 @@ class SpeechServer(speechserver.SpeechServer):
             pass
         else:
             name = acss_family.get(speechserver.VoiceFamily.NAME)
-            if name != self._default_voice_name:
-                self._send_command(set_synthesis_voice, name)
-            
+            self._send_command(set_synthesis_voice, name)
+
+    def _debug_sd_values(self, prefix=""):
+        if debug.debugLevel > debug.LEVEL_INFO:
+            return
+
+        try:
+            sd_rate = self._send_command(self._client.get_rate)
+            sd_pitch = self._send_command(self._client.get_pitch)
+            sd_volume = self._send_command(self._client.get_volume)
+            sd_language = self._send_command(self._client.get_language)
+        except:
+            sd_rate = sd_pitch = sd_volume = sd_language = "(exception occurred)"
+
+        family = self._current_voice_properties.get(ACSS.FAMILY)
+
+        styles = {settings.PUNCTUATION_STYLE_NONE: "NONE",
+                  settings.PUNCTUATION_STYLE_SOME: "SOME",
+                  settings.PUNCTUATION_STYLE_MOST: "MOST",
+                  settings.PUNCTUATION_STYLE_ALL: "ALL"}
+
+        current = self._current_voice_properties
+        msg = "SPEECH DISPATCHER: %s\n" \
+              "ORCA rate %s, pitch %s, volume %s, language %s, punctuation: %s \n" \
+              "SD rate %s, pitch %s, volume %s, language %s" % \
+              (prefix,
+               self._current_voice_properties.get(ACSS.RATE),
+               self._current_voice_properties.get(ACSS.AVERAGE_PITCH),
+               self._current_voice_properties.get(ACSS.GAIN),
+               self._get_language_and_dialect(family)[0],
+               styles.get(_settingsManager.getSetting("verbalizePunctuationStyle")),
+               sd_rate,
+               sd_pitch,
+               sd_volume,
+               sd_language)
+        debug.println(debug.LEVEL_INFO, msg, True)
+
     def _apply_acss(self, acss):
         if acss is None:
             acss = settings.voices[settings.DEFAULT_VOICE]
@@ -241,11 +310,13 @@ class SpeechServer(speechserver.SpeechServer):
             elif acss_property == ACSS.AVERAGE_PITCH:
                 method(5.0)
                 current[acss_property] = 5.0
-            elif acss_property == ACSS.FAMILY \
-                    and acss == settings.voices[settings.DEFAULT_VOICE]:
-                # We need to explicitly reset (at least) the family.
-                # See bgo#626072.
-                #
+            elif acss_property == ACSS.GAIN:
+                method(10)
+                current[acss_property] = 5.0
+            elif acss_property == ACSS.RATE:
+                method(50)
+                current[acss_property] = 5.0
+            elif acss_property == ACSS.FAMILY:
                 method({})
                 current[acss_property] = {}
 
@@ -262,6 +333,10 @@ class SpeechServer(speechserver.SpeechServer):
         Returns a text string with the punctuation symbols adjusted accordingly.
         """
 
+        style = _settingsManager.getSetting("verbalizePunctuationStyle")
+        if style == settings.PUNCTUATION_STYLE_NONE:
+            return oldText
+
         spokenEllipsis = messages.SPOKEN_ELLIPSIS + " "
         newText = re.sub(ELLIPSIS, spokenEllipsis, oldText)
         symbols = set(re.findall(PUNCTUATION, newText))
@@ -289,6 +364,44 @@ class SpeechServer(speechserver.SpeechServer):
     def _speak(self, text, acss, **kwargs):
         if isinstance(text, ACSS):
             text = ''
+
+        # Mark beginning of words with U+E000 (private use) and record the
+        # string offsets
+        # Note: we need to do this before disturbing the text offsets
+        # Note2: we assume that text mangling below leave U+E000 untouched
+        last_begin = None
+        last_end = None
+        marks_offsets = []
+        marks_endoffsets = []
+        marked_text = ""
+
+        for i in range(len(text)):
+            c = text[i]
+            if c == '\ue000':
+                # Original text already contains U+E000. But syntheses will not
+                # know what to do of it anyway, so discard it
+                continue
+
+            if not c.isspace() and last_begin == None:
+                # Word begin
+                marked_text += '\ue000'
+                last_begin = i
+
+            if c.isspace() and last_begin != None:
+                # Word end, add a mark
+                marks_offsets.append(last_begin)
+                marks_endoffsets.append(i)
+                last_begin = None
+
+            marked_text += c
+
+        if last_begin != None:
+            # Finished with a word
+            marks_offsets.append(last_begin)
+            marks_endoffsets.append(i + 1)
+
+        text = marked_text
+
         text = self.__addVerbalizedPunctuation(text)
         if orca_state.activeScript:
             text = orca_state.activeScript.\
@@ -304,8 +417,38 @@ class SpeechServer(speechserver.SpeechServer):
         #
         text = text.replace('\n.', '\n')
 
+        # Transcribe to SSML, translating U+E000 into marks
+        # Note: we need to do this after all mangling otherwise the ssml markup
+        # would get mangled too
+        ssml = "<speak>"
+        i = 0
+        for c in text:
+            if c == '\ue000':
+                if i >= len(marks_offsets):
+                    # This is really not supposed to happen
+                    msg = "%uth U+E000 does not have corresponding index" % i
+                    debug.println(debug.LEVEL_WARNING, msg, True)
+                else:
+                    ssml += '<mark name="%u:%u"/>' % (marks_offsets[i], marks_endoffsets[i])
+                i += 1
+            # Disable for now, until speech dispatcher properly parses them (version 0.8.9 or later)
+            #elif c == '"':
+            #  ssml += '&quot;'
+            #elif c == "'":
+            #  ssml += '&apos;'
+            elif c == '<':
+              ssml += '&lt;'
+            elif c == '>':
+              ssml += '&gt;'
+            elif c == '&':
+              ssml += '&amp;'
+            else:
+              ssml += c
+        ssml += "</speak>"
+
         self._apply_acss(acss)
-        self._send_command(self._client.speak, text, **kwargs)
+        self._debug_sd_values("Speaking '%s' " % ssml)
+        self._send_command(self._client.speak, ssml, **kwargs)
 
     def _say_all(self, iterator, orca_callback):
         """Process another sayAll chunk.
@@ -325,11 +468,21 @@ class SpeechServer(speechserver.SpeechServer):
                 t = self._CALLBACK_TYPE_MAP[callbackType]
                 if t == speechserver.SayAllContext.PROGRESS:
                     if index_mark:
-                        context.currentOffset = int(index_mark)
+                        index = index_mark.split(':')
+                        if len(index) >= 2:
+                            start, end = index[0:2]
+                            context.currentOffset = context.startOffset + int(start)
+                            context.currentEndOffset = context.startOffset + int(end)
+                            msg = "SPEECH DISPATCHER: Got mark %d:%d / %d-%d" % \
+                                (context.currentOffset, context.currentEndOffset, \
+                                 context.startOffset, context.endOffset)
+                            debug.println(debug.LEVEL_INFO, msg, True)
                     else:
                         context.currentOffset = context.startOffset
+                        context.currentEndOffset = None
                 elif t == speechserver.SayAllContext.COMPLETED:
                     context.currentOffset = context.endOffset
+                    context.currentEndOffset = None
                 GLib.idle_add(orca_callback, context, t)
                 if t == speechserver.SayAllContext.COMPLETED:
                     GLib.idle_add(self._say_all, iterator, orca_callback)
@@ -348,9 +501,8 @@ class SpeechServer(speechserver.SpeechServer):
         except KeyError:
             rate = 50
         acss[ACSS.RATE] = max(0, min(99, rate + delta))
-        debug.println(debug.LEVEL_CONFIGURATION,
-                      "Speech rate is now %d" % rate)
-
+        msg = 'SPEECH DISPATCHER: Rate set to %d' % rate
+        debug.println(debug.LEVEL_INFO, msg, True)
         self.speak(decrease and messages.SPEECH_SLOWER \
                    or messages.SPEECH_FASTER, acss=acss)
 
@@ -362,12 +514,24 @@ class SpeechServer(speechserver.SpeechServer):
         except KeyError:
             pitch = 5
         acss[ACSS.AVERAGE_PITCH] = max(0, min(9, pitch + delta))
-        debug.println(debug.LEVEL_CONFIGURATION,
-                      "Speech pitch is now %d" % pitch)
-
+        msg = 'SPEECH DISPATCHER: Pitch set to %d' % pitch
+        debug.println(debug.LEVEL_INFO, msg, True)
         self.speak(decrease and messages.SPEECH_LOWER \
                    or messages.SPEECH_HIGHER, acss=acss)
 
+    def _change_default_speech_volume(self, step, decrease=False):
+        acss = settings.voices[settings.DEFAULT_VOICE]
+        delta = step * (decrease and -1 or +1)
+        try:
+            volume = acss[ACSS.GAIN]
+        except KeyError:
+            volume = 10
+        acss[ACSS.GAIN] = max(0, min(9, volume + delta))
+        msg = 'SPEECH DISPATCHER: Volume set to %d' % volume
+        debug.println(debug.LEVEL_INFO, msg, True)
+        self.speak(decrease and messages.SPEECH_SOFTER \
+                   or messages.SPEECH_LOUDER, acss=acss)
+
     def getInfo(self):
         return [self._SERVER_NAMES.get(self._id, self._id), self._id]
 
@@ -377,11 +541,11 @@ class SpeechServer(speechserver.SpeechServer):
         from locale import getlocale, LC_MESSAGES
         locale = getlocale(LC_MESSAGES)[0]
         if locale is None or locale == 'C':
-            lang = None
-            dialect = None
+            locale_language = None
         else:
-            lang, dialect = locale.split('_')
-        voices = ((self._default_voice_name, lang, None),)
+            locale_lang, locale_dialect = locale.split('_')
+            locale_language = locale_lang + '-' + locale_dialect
+        voices = ()
         try:
             # This command is not available with older SD versions.
             list_synthesis_voices = self._client.list_synthesis_voices
@@ -392,12 +556,33 @@ class SpeechServer(speechserver.SpeechServer):
                 voices += self._send_command(list_synthesis_voices)
             except:
                 pass
-        families = [speechserver.VoiceFamily({ \
+
+        default_lang = ""
+        if locale_language:
+            # Check whether how it appears in the server list
+            for name, lang, variant in voices:
+                if lang == locale_language:
+                    default_lang = locale_language
+                    break
+            if not default_lang:
+                for name, lang, variant in voices:
+                    if lang == locale_lang:
+                        default_lang = locale_lang
+            if not default_lang:
+                default_lang = locale_language
+
+        voices = ((self._default_voice_name, default_lang, None),) + voices
+
+        families = []
+        for name, lang, variant in voices:
+
+            families.append(speechserver.VoiceFamily({ \
               speechserver.VoiceFamily.NAME: name,
               #speechserver.VoiceFamily.GENDER: speechserver.VoiceFamily.MALE,
-              speechserver.VoiceFamily.DIALECT: dialect,
-              speechserver.VoiceFamily.LOCALE: lang})
-                    for name, lang, dialect in voices]
+              speechserver.VoiceFamily.LANG: lang.partition("-")[0],
+              speechserver.VoiceFamily.DIALECT: lang.partition("-")[2],
+              speechserver.VoiceFamily.VARIANT: variant}))
+
         return families
 
     def speak(self, text=None, acss=None, interrupt=True):
@@ -426,12 +611,8 @@ class SpeechServer(speechserver.SpeechServer):
 
     def speakCharacter(self, character, acss=None):
         self._apply_acss(acss)
-        if character == '\n':
-            self._send_command(self._client.sound_icon, 'end-of-line')
-            return
-
         name = chnames.getCharacterName(character)
-        if not name:
+        if not name or name == character:
             self._send_command(self._client.char, character)
             return
 
@@ -440,12 +621,7 @@ class SpeechServer(speechserver.SpeechServer):
                 utilities.adjustForPronunciation(name)
         self.speak(name, acss)
 
-    def speakKeyEvent(self, event):
-        if event.isPrintableKey() and event.event_string.isupper():
-            acss = settings.voices[settings.UPPERCASE_VOICE]
-        else:
-            acss = ACSS(settings.voices[settings.DEFAULT_VOICE])
-
+    def speakKeyEvent(self, event, acss=None):
         event_string = event.getKeyName()
         if orca_state.activeScript:
             event_string = orca_state.activeScript.\
@@ -468,6 +644,12 @@ class SpeechServer(speechserver.SpeechServer):
     def decreaseSpeechPitch(self, step=0.5):
         self._change_default_speech_pitch(step, decrease=True)
 
+    def increaseSpeechVolume(self, step=0.5):
+        self._change_default_speech_volume(step)
+
+    def decreaseSpeechVolume(self, step=0.5):
+        self._change_default_speech_volume(step, decrease=True)
+
     def stop(self):
         self._cancel()
 
